from google.cloud import bigquery

from spintop_deploy.pub_sub_router import BigQueryJobHandler, BigQueryAuditLogMessage, PubSubRouter
from spintop.models.serialization import get_json_serializer

def bigquery_audit_handler(spintop_client, bq_client=None):
    
    if bq_client is None:
        bq_client = bigquery.Client()

    serializer = get_json_serializer()
    @BigQueryJobHandler
    def handle_bq_message(message: BigQueryAuditLogMessage):
        if message.table_id == 'spintop_metadata':
            table_identifier = f'`{message.project_id}`.`{message.dataset_id}`.`{message.table_id}`'
            metadata_result = bq_client.query(f'SELECT * FROM {table_identifier}').result()
            metadata = None

            for index, row in enumerate(metadata_result):
                # Only one row expected.
                if index > 0:
                    raise ValueError(f'Metadata returned more rows than expected (1)')

                metadata = dict(row)
            
            if not metadata:
                raise ValueError('Unable to retrieve metadata (no rows returned)')

            metadata = serializer.serialize(metadata)
            spintop_client.update_spintop_metadata(metadata)

            return True
        else:
            return False

    router = PubSubRouter([handle_bq_message])
    return router.request_handler()



from json import dumps
from google.cloud import tasks_v2 as tasks

class GoogleCloudTaskQueue(object):
    def __init__(self, project_id, gcp_region, gcp_queue_name, impersonate_service_account=None):
        self.client = tasks.CloudTasksClient()
        self.queue_path = self.client.queue_path(project_id, gcp_region, gcp_queue_name)
        self.impersonate_service_account = impersonate_service_account

    def enqueue_http_post(self, url, json=None):
        """queue_path = projects/PROJECT_ID/location/LOCATION_ID/queues/QUEUE_ID"""
        task = {
            "http_request": {  # Specify the type of request.
                "http_method": tasks.HttpMethod.POST,
                "url": url  # The full url path that the task will be sent to.
            }
        }

        if self.impersonate_service_account:
            task["http_request"]['oidc_token'] = {"service_account_email": self.impersonate_service_account}

        if json:
            task["http_request"]["headers"] = {"Content-type": "application/json"}
            task["http_request"]["body"] = dumps(json).encode()

        response = self.client.create_task(request={"parent": self.queue_path, "task": task})
        print("Created task {}".format(response.name))


import os
import time
import platform

from pprint import pprint


from flask import Flask, request
from spintop import __version__
from spintop.logs import setup_logging
from spintop.api_client import SpintopAPIError

from spintop_deploy.logs import _logger
from spintop_deploy.env import SpintopDeployEnv

from .bigquery_audit_handler import bigquery_audit_handler

logger = _logger('cloud-agent')

def create_app(init_env=None):

    app = Flask(__name__)
    bootstrap_env = SpintopDeployEnv(init_env)
    modules = bootstrap_env.deployment_provider_modules()
    modules.load(setup_logging)

    param_provider = modules['param_provider']

    if not bootstrap_env.SPINTOP_M2M_SECRET_KEY:
        bootstrap_env.SPINTOP_M2M_SECRET_KEY  = param_provider.get_environ_or_param_value('SPINTOP_M2M_SECRET_KEY', default=None)

    spintop_client = bootstrap_env.spintop_factory()
    modules.update(spintop_client=spintop_client)

    logger.info(spintop_client.auth.describe())

    @app.route('/', methods=['GET'])
    def index():
        
        status = None
        remote_status = None
        authentication_ok = True
        authorization_ok = True
        try:
            org_info = spintop_client.get_org_info()
        except SpintopAPIError as e:
            status = 'ERROR'
            remote_status = 'OK'

            org_info = e.resp.json()
            if e.resp.status_code == 401:
                authentication_ok = False
                authorization_ok = False
            elif e.resp.status_code == 403:
                authorization_ok = False

        except Exception as e:
            status = 'ERROR'
            remote_status = 'ERROR'

            org_info = 'Unable to retrieve org info: ' + str(e)
        
        return {
            'status': 'OK' if status is None else status,
            'version': __version__,
            'org_info': org_info,
            'remote': spintop_client.api_url,
            'remote_status': 'OK' if remote_status is None else remote_status,
            'remote_authentication': 'OK' if authentication_ok else 'ERROR',
            'remote_authorization': 'OK' if authorization_ok else 'ERROR'
        }, 200 if status is None else 500 # Config error

    handler = modules.load(bigquery_audit_handler)

    @app.route('/', methods=['POST'])
    def notification():
        try:
            data = request.json
            print('POST content:')
            print(data)
        except Exception as e:
            print('Unable to get JSON: ' + str(e))
            raise

        handled = handler(data)
        
        return { 'handled': handled }, 200
    
    return app


#!/usr/bin/env python3

import simplejson as json
from io import BytesIO
from tempfile import TemporaryFile

from google.cloud import bigquery
from google.cloud.bigquery.job import SourceFormat
from google.cloud.bigquery import Dataset, WriteDisposition
from google.cloud.bigquery import LoadJobConfig
from google.api_core import exceptions

from spintop.logs import _logger

logger = _logger('bigquery')

class BigQueryJobBuilder(object):
    def __init__(self, dataset_id, table_name, truncate=False, project_id=None, client=None, dataset=None):
        self.dataset_id = dataset_id
        self.table_name = table_name

        self.truncate = truncate
        self.schema = None

        if client is None:
            client = bigquery.Client(project=project_id)

        if dataset is None:
            dataset = client.dataset(dataset_id)
        
        self.dataset = dataset
        self.client = client

    def _build_job_from_stream(self, file_stream, load_config):
        dataset_ref = self.dataset
        table_ref = dataset_ref.table(self.table_name)

        file_ref = file_stream
        data_bytes = file_ref.tell()
        file_ref.seek(0)
        load_job = self.client.load_table_from_file(
            file_ref, table_ref,
            job_config=load_config
        )
        logger.info(f'Created job {load_job.job_id} of {data_bytes/1000000} MBytes')
        return load_job

class BigQueryDictLoadJobBuilder(BigQueryJobBuilder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = TemporaryFile(mode='w+b')

    def add_row(self, row_dict):
        dat = bytes(json.dumps(row_dict) + '\n', 'UTF-8')
        self.data.write(dat)

    def add_rows(self, rows):
        for row in rows:
            self.add_row(row)
    
    @property
    def is_empty(self):
        return self.data.tell() == 0
 
    def build_job(self):
        load_config = LoadJobConfig()
        if self.schema is None:
            # No schema. Use Bigquery auto schema
            load_config.autodetect = True
        else:
            load_config.schema = self.schema

        load_config.source_format = SourceFormat.NEWLINE_DELIMITED_JSON
    
        if self.truncate:
            load_config.write_disposition = WriteDisposition.WRITE_TRUNCATE
        else:
            # Allow field addition or relaxation if not truncate.
            load_config.schema_update_options = [
                bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
                bigquery.SchemaUpdateOption.ALLOW_FIELD_RELAXATION
            ]

        return self._build_job_from_stream(self.data, load_config)

class BigQueryParquetLoadJobBuilder(BigQueryDictLoadJobBuilder):

    def set_bytes_stream(self, data):
        self.data = data

    def build_job(self):
        load_config = LoadJobConfig()

        load_config.source_format = SourceFormat.PARQUET
    
        if self.truncate:
            load_config.write_disposition = WriteDisposition.WRITE_TRUNCATE
        else:
            # Allow field addition or relaxation if not truncate.
            load_config.schema_update_options = [
                bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
                bigquery.SchemaUpdateOption.ALLOW_FIELD_RELAXATION
            ]

        return self._build_job_from_stream(self.data, load_config)

    


import os
import sys
import yaml
import tempfile
from pprint import pprint
import subprocess

import dbt.main

from spintop.api_client.models import ManyOrganizations

class BigqueryOutputFactory(object):
    def __init__(self, project_id, keyfile=None, threads=40, impersonate_service_account=None):
        if keyfile is None:
            keyfile = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', None)

        if keyfile:
            keyfile = os.path.abspath(keyfile)
            method = 'service-account'
        else:
            method = 'oauth'
        
        self.keyfile = keyfile
        self.project_id = project_id
        self.method = method
        self.threads = threads
        self.impersonate_service_account = impersonate_service_account

    @classmethod
    def from_env(cls, env):
        kwargs = dict(
            project_id=env.GOOGLE_CLOUD_PROJECT
        )

        keyfile = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', None)

        if not keyfile: 
            if env.IMPERSONATE_SERVICE_ACCOUNT:
                kwargs['impersonate_service_account'] = env.IMPERSONATE_SERVICE_ACCOUNT
        else:
            kwargs['keyfile'] = keyfile
        
        return cls(**kwargs)

    def create(self, org_data, env_name):
        dataset = org_data.env_dataset_name(env_name)
        return dict(
            type='bigquery',
            method=self.method,
            threads=self.threads,
            keyfile=self.keyfile,
            project=self.project_id,
            impersonate_service_account=self.impersonate_service_account,
            dataset=dataset
        )

class SpintopOrgProfileFactory(object):
    def __init__(self, output_factory):
        self.output_factory = output_factory

    def create_profile(self, org_data):
        org_key = org_data.key
        return {
            org_key: dict(
                outputs={
                    env.name: self.output_factory.create(org_data, env.name) for env in org_data.envs
                }
            )
        }

def create_all_profiles(all_profiles):
    profiles = {}
    for p in all_profiles:
        profiles.update(p)
    return profiles

def dbt_run(project_dir, profile_dir, org_key, env_data, dbt_args=[], debug=False):
    vars_string = yaml.safe_dump(env_data.vars, default_flow_style=True).strip()
    print(f'Running DBT in project-dir={project_dir} with target={env_data.name}. Vars are:')
    print(vars_string)
    args = [
        'run',
        '--project-dir', project_dir,
        '--profiles-dir', profile_dir,
        '--profile', org_key,
        '--target', env_data.name,
        '--vars', vars_string
    ] + list(dbt_args)

    if debug:
        args.insert(0, '-d')
    
    # print('Would run')
    # pprint(args)
    p = subprocess.Popen([sys.executable, '-m', __name__] + args)
    out, err = p.communicate()

    if p.returncode != 0:
        raise RuntimeError('Error while executing dbt run.')


class DBTInterface(object):
    def __init__(self, project_dir, output_factory, temp_dir=None):
        self.project_dir = project_dir
        self.output_factory = output_factory

        if temp_dir is None:
            temp_dir = tempfile.mkdtemp(suffix='dbt-interface')

        self.temp_dir = temp_dir
        self.orgs = None

    @property
    def profiles_file(self):
        return os.path.join(self.temp_dir, 'profiles.yml')

    def load_orgs(self, orgs):
        self.orgs = ManyOrganizations(orgs=list(orgs))
        factory = SpintopOrgProfileFactory(self.output_factory)
        all_profiles = [factory.create_profile(org) for org in orgs]

        profiles = create_all_profiles(all_profiles)
        with open(self.profiles_file, 'w+') as profiles_file:
            yaml.safe_dump(profiles, profiles_file)

    def run(self, org_key, env_name, dbt_args=[], debug=False):
        env = self.orgs.find_org(org_key).find_env(env_name)
        return dbt_run(
            project_dir=self.project_dir, 
            profile_dir=self.temp_dir,
            org_key=org_key, 
            env_data=env, 
            dbt_args=dbt_args, 
            debug=debug
        )

if __name__ == '__main__':
    print('CWD: ' + os.getcwd())
    dbt.main.main()

    




import re
from collections import namedtuple
from typing import List, Dict, Any, Optional
import datetime as dt


from spintop.models import BaseDataClass

from .models import OrgDataEnv, EnvDataspecs, DimensionSpec, TableSpec

def create_env_from_spintop_metadata(spintop_metadata):
    metadata_factories = {
        'dbt': create_env_from_dbt_metadata
    }

    meta_style = spintop_metadata.get('style', None)

    if not meta_style:
        raise ValueError('Metadata is missing "style" field.')

    factory = metadata_factories.get(meta_style, None)

    if not factory:
        raise ValueError(f'Metadata style {meta_style!r} is not supported.')

    return factory(spintop_metadata)

class Versions(BaseDataClass):
    spintop: str
    spintop_da: str

class SpintopDAVariables(BaseDataClass):
    step_nodes: dict
    transitions: dict

    def as_vars(self):
        all_vars = {}
        fields = list()
        for field in self.fields():
            all_vars['spintop_da:' + field.field_name_] = field.get_value(self)
        return all_vars

class BigQueryTableSchema(BaseDataClass):
    table_catalog: str
    table_schema: str
    table_name: str
    table_type: str
    is_insertable_into: str
    is_typed: str
    creation_time: str

class BigQueryColumnSchema(BaseDataClass):
    table_catalog: str
    table_schema: str
    table_name: str
    column_name: str
    ordinal_position: int = None
    is_nullable: str = None
    data_type: str
    is_generated: str = None
    generation_expression: str = None
    is_stored: str = None
    is_hidden: str = None
    is_updatable: str = None
    is_system_defined: str = None
    is_partitioning_column: str = None
    clustering_ordinal_position: int = None

class DBTStyleMetadata(BaseDataClass):
    dataset_id: str
    project_id: str
    target: Dict[str, Optional[Any]] = dict
    last_run_at: str
    dbt_graph: str ## It's json, but so complex that the parser cannot handle it. Keep as str.
    versions: Versions
    spintop_da_vars: SpintopDAVariables
    dataset_tables: List[BigQueryTableSchema]
    ref_columns: List[BigQueryColumnSchema]

def create_env_from_dbt_metadata(metadata):
    schema = DBTStyleMetadata.get_schema('tabular')()
    metadata = schema.load(metadata)

    columns = index_columns_by_table_name_dot_column_name(metadata.ref_columns)

    dbt_vars = {
        'source_project': metadata.project_id,
        'source_dataset': metadata.dataset_id
    }
    dbt_vars.update(metadata.spintop_da_vars.as_vars())

    env = OrgDataEnv(
        name=metadata.target['name'],
        vars=dbt_vars
    )

    dataspecs = EnvDataspecs(
        project_id=metadata.project_id,
        dataset_id=metadata.dataset_id
    )

    transitions_and_nodes_user_fields = (
        metadata.spintop_da_vars.transitions.get('keep_steps_entered_fields', []) + 
        metadata.spintop_da_vars.transitions.get('keep_previous_steps_entered_fields', []) +
        metadata.spintop_da_vars.step_nodes.get('super_group_by_lc_types', []) +
        metadata.spintop_da_vars.step_nodes.get('group_by_lc_types', [])
    )



    dataspecs.production_nodes = TableSpec(
        custom_dimensions=list(
            create_user_fields_dimensions(
                'production_nodes', 
                transitions_and_nodes_user_fields,
                columns
            )
        )
    )

    env.dataspecs = dataspecs
    return env

def index_columns_by_table_name_dot_column_name(ref_columns):
    indexed = {}
    for col in ref_columns:
        indexed[f'{col.table_name}.{col.column_name}'] = col
    return indexed

def create_user_fields_dimensions(table_name, user_fields, columns):
    for user_field in user_fields:
        field_index = f'{table_name}.{user_field}'

        if field_index not in columns:
            pass # No dimensions.
        else:
            col = columns[field_index]
            data_type = col.data_type
            yield from expand_data_type(user_field, data_type)


StructSubType = namedtuple('StructSubType', ['field_name', 'data_type'])
struct_regex = re.compile(r'STRUCT<([\s\S]*)>')
fields_regex = re.compile(r'(?:(\w+) (STRUCT<[\s\S]*>|[\w]+))')
def expand_data_type(field_name, data_type, parent_fields=[]):

    struct_data_type_match = struct_regex.match(data_type.strip())
    if struct_data_type_match:
        # It's a struct.
        fields_str = struct_data_type_match.group(1)
        for match in fields_regex.finditer(fields_str):
            sub_field_name = match.group(1)
            data_type =  match.group(2)
            yield from expand_data_type(sub_field_name, data_type, parent_fields=parent_fields + [field_name])

    elif field_name != '_type':
        yield DimensionSpec(
            name='.'.join(parent_fields + [field_name]),
            field_type=data_type.lower()
        )



import os
import dotenv
dotenv.load_dotenv()

from .handler import create_app

if __name__ == "__main__":
    app = create_app()
    app.run(debug=True, host='0.0.0.0', port=int(os.environ.get('PORT', 5080)))


from typing import List
from enum import Enum
from datetime import datetime

from spintop.models import (
    BaseDataClass,
    PersistenceIDRecord,
    OutcomeData,
    VersionIDRecord
)

class LCTypes():
    DUT = 'dut'
    STATION = 'station'
    OPERATOR = 'operator'
    SOFTWARE = 'software'
    STEP = 'step'

class LCUnit(VersionIDRecord):
    lc_type: str = ''

    @classmethod
    def from_version_id(cls, version_id, lc_type):
        return cls(
            id=version_id.id,
            version=version_id.version,
            lc_type=lc_type
        )

class LCStep(PersistenceIDRecord):
    """
    Steps represent a distinct production event. 
    
    Each row is a particular event with a date and time, an outcome and multiple associated lifecycle units.

    ## Lifecycle Unit

    Each step references many lifecycle units. Only one of these units, 

    ## Lifecycle Type

    This is the unit type that will be tracked throughout the production line. 
    Say a certain step represents a system test at the step 'system test', executed
    by operator 'Mike'. The system ID is 'SYS01' and a board with ID 'BOARD01' is 
    assembled in this system. The step units should be:

    units:
    - LCUnit(lc_type='system', id='SYS01')
    - LCUnit(lc_type='board', id='BOARD01')
    - LCUnit(lc_type='step', id='system test')
    - LCUnit(lc_type='operator', id='Mike')
    
    And, because the DUT is the system, the lc_type attribute of the step itself
    should be 'system'.
    """
    
    fields_docs_ = dict(
        outcome=OutcomeData.__doc__,
        duration="""
    The duration in seconds of this step. Use None/null if unknown.
        """,
        lc_type="""
    This represents the lifecycle type of this main lifecycle unit of this step. 
    See the lifecycle type section of LCStep documentation.
        """,
        units="""
    The list of lifecycle units associated with this step. One of those is the 'main'
    device under test (DUT) as determined by the lc_type attribute.
        """

    )
    fields_docs_.update(PersistenceIDRecord.fields_docs_)

    outcome: OutcomeData = None
    duration: float = None
    lc_type: str = None # The lifecycle of this step. Linked with the units lc_types.
    units: List[LCUnit] = list
    filename: str = None
    pipeline_uuid: str = None
    
    def __post_init__(self):
        super().__post_init__()
        if self.outcome is not None:
            self.outcome = OutcomeData.create(self.outcome)



import os
import json
import base64
import platform

from flask import Blueprint, request

import prefect
from prefect import Flow

from .agent import ServerlessAgent

class SpintopServerlessPrefectContext(object):
    def __init__(self, prefect_project, storage, agent_name='sls', id_map_encoded=None, param_provider=None):
        if param_provider:
            load_secrets(param_provider)

        self.prefect_project = prefect_project
        self.storage = storage
        self.agent_name = agent_name
        self.id_map = {}
        self._flows = {}

        if id_map_encoded:
            self.decode_id_map(id_map_encoded)

    def add_flow(self, *flows):
        for flow in flows:
            flow.storage = self.storage
            self.storage.add_flow(flow)
            self._flows[flow.name] = flow
        return flow

    def build(self):
        self.storage.build()

    def register(self):
        self.build()
        for flow_name, flow in self._flows.items():
            flow_id = flow.register(project_name=self.prefect_project, build=False)
            self.id_map[flow_name] = flow_id
        return self.id_map
    
    def encode_id_map(self):
        json_data = json.dumps(self.id_map)
        base64_bytes = base64.b64encode(json_data.encode('utf-8'))
        return base64_bytes.decode('ascii')

    def decode_id_map(self, map_encoded):
        base64_bytes = map_encoded.encode('ascii')
        json_data = base64.b64decode(base64_bytes).decode('utf-8')
        self.id_map = json.loads(json_data)
        return self.id_map

    def run(self, flow_name, parameters={}, **kwargs):
        flow_id = self.id_map.get(flow_name, None)
        if flow_id is None:
            raise ValueError(f'No flow with name {flow_name}')
        
        ephemeral_agent = ServerlessAgent(name=self.agent_name, labels=[flow_id], show_flow_logs=True)
        ephemeral_agent.run_flow_id(flow_id, labels=[flow_id], parameters=parameters)

    def flask_blueprint(self):
        from flask import Blueprint, request

        blueprint = Blueprint('prefect_flows', __name__)
        
        links = {}
        for flow_name, flow_id in self.id_map.items():
            blueprint.route(f'/{flow_name}', methods=['POST', 'GET'])(
                self.create_route(flow_name, flow_id)
            )
            links[flow_name] = flow_id


        @blueprint.route('/', methods=['GET'])
        def index():
            return {
                'status': 'OK',
                'flows': links
            }

        return blueprint

    def create_route(self, flow_name, flow_id):
        def _route():
            try:
                parameters = request.json
            except Exception as e:
                parameters = {}

            ephemeral_agent = ServerlessAgent(name=self.agent_name, labels=[flow_id], show_flow_logs=True)
            ephemeral_agent.run_flow_id(flow_id, labels=[flow_id], parameters=parameters)
            return 'OK', 200
        
        _route.__name__ = flow_name
        return _route

def load_secrets(param_provider):
    os.environ['PREFECT__CLOUD__AGENT__AUTH_TOKEN'] = param_provider.get_environ_or_param_value('PREFECT__CLOUD__AGENT__AUTH_TOKEN', default='')
    prefect.context.config.cloud.agent.auth_token = os.environ["PREFECT__CLOUD__AGENT__AUTH_TOKEN"]
    prefect.config.cloud.agent.auth_token = os.environ["PREFECT__CLOUD__AGENT__AUTH_TOKEN"]

    os.environ['PREFECT__CLOUD__AUTH_TOKEN'] = param_provider.get_environ_or_param_value('PREFECT__CLOUD__AUTH_TOKEN', default='')
    prefect.context.config.cloud.auth_token = os.environ["PREFECT__CLOUD__AUTH_TOKEN"]
    prefect.config.cloud.auth_token = os.environ["PREFECT__CLOUD__AUTH_TOKEN"]